;; Planar7:
;; 1) allows for any range of initialization in the function start.  That is
;;    the planarity, ect. of the graph is checked and set at initialization so
;;    that positive potentials may be assigned during initialization
;;
;; 2) allows the user to decide wheather the groups should be combined or not.
;;    This is done by seting partition to either: T to combine groups or F to
;;    not combine them.
;;
;; 3) allows the user to determine a 'sweep strategy' be seting the value of 
;;    the variables alternate and upsweep.  
;;
;; ****************************************************************************
;; Simplest approach to planarity testing - one calculation, no attempt
;; to optimize placement of communicating processors, but this version
;; is arranged for graphs which are far from complete.
;;
;; Non-trivial edges are partitioned into independent classes so as to
;; permit a reasonably parallel, Gauss-Seidel style of iteration.
;; One cycle requires at most O(n) parallel steps, for a graph of n vertices.
;;
;; Interaction processors are placed in consecutive cube addresses.
;; There's no attempt to separate them according to edge classes.
;; Edge representatives are divided:
;;
;;   |0  - ntrve's -  |EL       |EH - trve's - |
;;
;; Initially, the program constructs edges for the complete graph,
;; then rearranges; this could be changed.
;;
;; Interactions are stored in vpinter, only as many as needed:
;;
;;   |0    -  interactions    -  |IL . . .
;;
;; We set up a representation for exactly the graph given.
;; This implements Sibert's modification of Takefuji and Lee, which
;; avoids conflict when both Uup and Udn would become > 0, decreasing
;; the smaller of the two.  See loadgp.lisp for functions which
;; set up graphs either from files describing them or randomly.

;; Uses a separate vp set for interaction processors

(in-package '*lisp)

(defvar program nil)
(setq program 'planar7)

;; Indicate that this program wants edges and interactions partitioned
(setq partition t)
;; Indicate that this program wants to alternate its sweep paths
(defvar alternate t)
;; Indicate that this program wants to do it's first sweep in the up direction
(defvar upsweep t)
;; These are used in the cycle program to control altern. sweep capabilities
(defvar ex 0)
(defvar incr)

(defconstant NS 15)	;; Size of vertex number (bits)
(defconstant NC 16)	;; Size of (signed) class number (bits)
(defconstant 1k 1024)	;; For convenience

(*proclaim '(type fixnum nv ce Ninter EL EH DSZ IL))

(defvar nv 4)          ;; Number of vertices in graph
(defvar ce 6)          ;; Number of edges in current graph
(defvar Ninter 16384)  ;; Size of vp set for interactions
(defvar EL 0)          ;; Next position for non-trivial edge = no. of same
(defvar EH 8191)       ;; First position for trivial edge
(defvar DSZ 16384)      ;; Default vp-set size
(defvar IL 0)          ;; Next position for interaction = no. of same

(def-vp-set vpinter (list Ninter))  ;; 1-D vp set for interactions

(*proclaim '(type Boolean-pvar edgep edgsg ntrve edgex intrp prsnt))

(*defvar edgep nil!!)	;; This processor represents an edge
(*defvar edgsg nil!!)	;; Segmentation pvar for edges
(*defvar ntrve nil!!)	;; This edge is non-trivial
(*defvar prsnt t!!)	;; When edgep, this edge is present in the graph

(*defvar intrp nil!! "i" vpinter)  ;; This processor manages an interaction

(*proclaim '(type (field-pvar NS) edga edgb)) ;; ntra ntrb ntrc ntrd

(*defvar edga (!! 0))      ;; When edgep, edge is [edga, edgb]
(*defvar edgb (!! 0))

(*proclaim '(type (signed-pvar NC) edgc))

(*defvar edgc (!! 0))		;; Edge class (among independent sets)

;; ntra, b, c, d retained for testing; they aren't really needed

;; (*defvar ntra (!! 0) "i" vpinter)  ;; When intrp, interaction is between
;; (*defvar ntrb (!! 0) "i" vpinter)  ;;  [ntra, ntrb] and [ntrc, ntrd],
;; (*defvar ntrc (!! 0) "i" vpinter)  ;; whence
;; (*defvar ntrd (!! 0) "i" vpinter)  ;;  0 <= ntra < ntrc < ntrb < ntrd < nv

(*proclaim '(type (field-pvar 21) self rnvz rpab rpcd))

(*defvar self (self-address!!))	;; For convenience
(*defvar rnvz (!! 0))		;; For rendezvous calculations

(*defvar rpab (!! 0) "i" vpinter)	;; When intrp, rep of [a,b]
(*defvar rpcd (!! 0) "i" vpinter)	;; When intrp, rep of [c,d]

(*proclaim '(type (field-pvar NS) clab clcd))

(*defvar clab (!! 0) "i" vpinter)	;; Class of rpab
(*defvar clcd (!! 0) "i" vpinter)	;; Class of rpcd

(*proclaim '(type (unsigned-pvar 2) Iud Vud))
(*defvar Iud (!! 0) "i" vpinter)  ;; determines up/down interactions at once
(*defvar Vud (!! 0))  ;;  contained both up/down voltages
                      ;;  0 bit = up, 1 bit = down

(*proclaim '(type (signed-pvar 24) cntx cnty ONE))

(*defvar cntx (!! 0))                    ;; For enumeration, etc.
(*defvar cnty (!! 0))                    ;; Working counter
(*defvar ONE  (!! 1) "(!! 1)" vpinter)   ;; For send-with-add

;; Pvar-locations for CM:send-with-s-add

(*proclaim '(type fixnum cntx-addr cnty-addr itgt-addr ONE-addr))

(defvar cntx-addr 0)  (defvar cnty-addr 0)  (defvar ONE-addr 0)
(defvar itgt-addr 0)

(*proclaim '(type (signed-pvar 24) isrc itgt))
(*defvar isrc (!! 0) "i" vpinter)  ;; inter. source: address to read Vud from
(*defvar itgt (!! 0) "i" vpinter)  ;; inter. target: address to send sum

(*proclaim '(type Boolean-pvar iactv))
(*defvar iactv Nil!! "i" vpinter) ;; boolean-pvar for interactions active 
                                  ;; during subsycle (used in cycle program)

;; Cold boot with 1-D vp set of size specified by two factors, default 8k

(defun setup (&optional (xs 128) (ys 128))
 (setq DSZ (* xs ys))
 (*cold-boot :initial-dimensions (list DSZ))
 (setq cntx-addr (pvar-location cntx))
 (setq cnty-addr (pvar-location cnty))
 (setq itgt-addr (pvar-location itgt))
 (setq ONE-addr  (pvar-location ONE))
 (list 'DEFAULT DSZ 'INTERACTIONS Ninter)
)

;; ******* Functions for graph and interaction construction *******

(defvar outcome 'SUCCESS)  ;; Result of graph construction

;; Set up edges for complete graph

(defun mk-edges ()
 ;; Clear important parts of existing representations
 (*set edgep nil!!)  (*set edgsg nil!!)  (*set ntrve nil!!)  (*set edga (!! 0))
 (setq outcome 'SUCCESS)
 ;; Make temporary vertices and begin base edge construction
 (*when (<!! self (!! nv)) ; Temporary vertex rep.
   (*set cntx (-!! (!! nv) (1+!! self)))     ; Number of edges originating here
   (*set rnvz (-!! (scan!! cntx '+!!) cntx)) ; First edge address
   (*pset :no-collisions self edga rnvz :notify edgsg)
                                 ; Originating vertex to first edge
 )
 ;; Fill in base edge representations for complete graph
 (*set edga (scan!! edga 'copy!! :segment-pvar edgsg))  ; Origins
 (*when (<!! edga (!! (1- nv)))   ; Real edge rep.
   (*set edgep t!!)  (setq ce (*sum (!! 1)))
   (*set edgb (+!! edga (scan!! (!! 1) '+!! :segment-pvar edgsg))) ; Dest.
   (*set ntrve (not!! (or!! (=!! edgb (1+!! edga))   ; Non-trivial edges
                            (and!! (zerop!! edga) (=!! edgb (!! (1- nv)))))))
 )
)

;; Separate trivial from non-trivial edges

(defun split-edges ()
 (*set edgep nil!!)
 (*when prsnt
   (setq ce (*sum (!! 1)))
   (format t "~D edges in this graph.~%" ce)
   (*if ntrve
      (progn (*set rnvz (enumerate!!))
             (setq EL (*sum (!! 1)))  (setq EH (- DSZ (- ce EL))))
      (*set rnvz (+!! (!! EH) (enumerate!!))))
   (*set rnvz (cube-from-grid-address!! rnvz))
   (*pset :no-collisions edga edga rnvz :notify edgep)
   (*pset :no-collisions edgb edgb rnvz)
 ) ; end *when prsnt
 (*set ntrve (<!! (self-address-grid!! (!! 0)) (!! EL)))
 (*set prsnt edgep)
)

;; Partition edges into indedpendent classes.  Note that vertices are
;;
;;	0    1    2    .  .  .    nv-3    nv-2    nv-1
;;
;; so that last class begins at nv-3.

(*proclaim '(type (signed-pvar NC) cl fa fb ma mb mc))

(*defvar cl (!! 0))	;; Fixed class number
(*defvar fa (!! 0))	;; Fixed left limit
(*defvar fb (!! 0))	;; Fixed right limit
(*defvar ma (!! 0))	;; Moving class number = left limit (or -1)
(*defvar mb (!! 0))	;; Moving, right limit

(defvar Ncl 0)		;; Number of non-empty classes
(defvar combine t)      ;; T if classes should be combined

;; Display for class calulation

(defun disp-class ()
  (let ((fin (1- nv)))
    (ppp cl :end fin)  (ppp ma :end fin)  (ppp mb :end fin)
  )
)


;; Systolic calculation to generate interactions, incorporates classes

(defun mk-inter ()
 (setq IL 0)
 (*with-vp-set vpinter  (*set intrp nil!!))  ;; Clear representation
 (*let ((mvga edga) (mvgb edgb) (mvgc edgc))
       (declare (type (field-pvar NS) mvga mvgb)
		(type (signed-pvar NC) mvgc))
   ;; grid address of moving edge rep for loop:
   (*set cntx (1+!! (self-address-grid!! (!! 0))))
   ;; To prevent garbage interactions:
   (*when (not!! ntrve) (*set mvga (!! 0)) (*set mvgb (!! 0)))
   (*when (not!! edgep) (*set edga (!! 0)) (*set edgb (!! 0)))
   ;; Note that wrap-around is harmless, as test will ignore these.
   ;; We leave (possibly many) processors active - unused or trivial edges
   (dotimes (ss (1- EL))
     (*news mvga mvga -1)  (*news mvgb mvgb -1)  (*news mvgc mvgc -1)
     (*when (<!! edga mvga edgb mvgb)  ;; Interaction found
        (*set rnvz (+!! (!! IL) (enumerate!!)))  ;; Interaction address
        (setq IL (+ IL (*sum (!! 1))))
        (if (zerop (mod ss 50))  ;; Progress report
          (format t "Step ~D:  ~D interactions now.~%" ss IL))
        (if (> IL Ninter)
          (progn
            (format t "***** No more space, ~D processors needed. *****~%" IL)
            (setq outcome 'FAILURE))
          (progn    ;; Fill in interactions - ntr's aren't needed
           ;; (*pset :no-collisions edga ntra rnvz :vp-set vpinter)
           ;; (*pset :no-collisions edgb ntrb rnvz :vp-set vpinter)
           ;; (*pset :no-collisions mvga ntrc rnvz :vp-set vpinter)
           ;; (*pset :no-collisions mvgb ntrd rnvz :vp-set vpinter)
            (*pset :no-collisions self rpab rnvz
                                   :notify intrp :vp-set vpinter)
	    (*pset :no-collisions edgc clab rnvz :vp-set vpinter)
            (*set cnty (cube-from-grid-address!! cntx))
            (*pset :no-collisions cnty rpcd rnvz :vp-set vpinter)
	    (*pset :no-collisions mvgc clcd rnvz :vp-set vpinter))
         )) ; end if, end *when
     (if (eq outcome 'FAILURE) (return-from mk-inter outcome))
     (*set cntx (1+!! cntx))
   )) ; end dotimes, end *let
 (format t "Finished at step ~D:  ~D interactions.~%" (- EL 2) IL)
 outcome
) ; end mk-inter

;; Neural net state and parameters

(*proclaim '(type Boolean-pvar Vup Vxp Vdn Vxn acte))

(*defvar Vup  nil!!)   ;; Up edge
(*defvar Vxp  nil!!)   ;; Best legal Up configuration so far
(*defvar Vdn  nil!!)   ;; Down edge
(*defvar Vxn  nil!!)   ;; Best legal Dn configuration so far
(*defvar acte nil!!)   ;; edgep and ntrve and prsnt (active edge rep)

(*proclaim '(type single-float-pvar Uup Udn Du PD))

(*defvar Uup (!! 0.0))  ;; Up activation
(*defvar Udn (!! 0.0))  ;; Down activation
(*defvar Du  (!! 0.0))  ;; Delta U (for both up and down in turn)
(*defvar PD  (!! 0.0))  ;; Up-down contribution to Du

(*proclaim '(type single-float A B DT A*DT))

(defvar A  2.0)       ;; Coefficient for up-down conflict
(defvar B  1.0)       ;; Coefficient for edge conflicts
(defvar DT 1.0E-5)    ;; Delta-t
(defvar A*DT 2.0E-5)  ;; Will be A*DT
(defvar nlegal '())     ;; new legal states since last change of A

(defvar cycles 0)     ;; Counts iterations of nueral algorithm
(defvar legal nil)    ;; Used to report conflicts, T iff last configuration
                      ;; is legal
(defvar more nil)     ;; Used to report potential for more edges, true iff
                      ;; Du > 0 somewhere
(defvar trvedges 0)   ;; Number of trivial edges
(defvar ntredges 0)   ;; Current number of non-trivial edges
(defvar totedges 0)   ;; Current total number of edges
(defvar maxedges 0)   ;; Maximum number of edges so far in legal graph

;; Random float!! between lo and hi

(*defun rf!! (lo hi)
 (+!! (*!! (random!! (!! 1000000)) (!! (/ (- hi lo) 1E6))) (!! lo))
)

;; Start neural-net calculation, after T&L, parameters are lower, upper
;; bounds for initial activations.  Handles trivial edges.

(defun start (&optional (Ul (- Dt)) (Uh 0))
 ;; backwards b/c cycle changes ex and incr
 (if (lcl:xor upsweep alternate) (progn (setq incr 1) (setq ex (1- Ncl)))
                             (progn (setq incr -1) (setq ex 0)))
 (setq cycles 0)
 (newa 2.0)
 (setq A*DT (* A DT))
 (*set acte (and!! edgep ntrve))
 (*when acte
   (*set Uup (rf!! Ul Uh))  (*set Udn (rf!! Ul Uh))
  )
;; this eliminates double edges that may occur on initialization
 (*when (and!! (plusp!! Uup) (plusp!! Udn))
    (*if (>=!! Uup Udn) (*set Udn (-!! Udn (!! A*DT)))
                        (*set Uup (-!! Uup (!! A*DT))))
 )
 (*set Vup (plusp!! Uup)) (*set Vdn (plusp!! Udn))
 (*when (and!! edgep (not!! ntrve))  ;; Trivial edges (all are present)
   (*set Vup (evenp!! edga)) (*set Vdn (not!! Vup))
   (setq trvedges (*sum (!! 1)))
 )
 (*when acte  ;; To count non-trivial edges
    (*when (or!! Vup Vdn) (setq ntredges (*sum (!! 1))))
 )
 (setq totedges (+ trvedges ntredges))
 (testplanar Vup Vdn)                    ;; Test for planarity, maximality
 (if legal (progn (*set Vxp Vup) (*set Vxn Vdn) (setq maxedges totedges))
           (progn (*set Vxp (not!! ntrve)) (*set Vxn (not!! ntrve)) (setq maxedges 0)))
 (status)
 (svpresnt "edges.list")
 )

(defun sum (l)  (cond ((null l) 0) (T (+ (car l) (sum (cdr l))))))

;; Perform one cycle of algorithm.
;; Call from default vp-set, all processors active

(defun cycle ()
 ;; self-adjust mechanism:
 (if (> (length nlegal) 4)  
     (if (> (/ (sum nlegal) (length nlegal)) .7) (newa (+ A .3))
     (if (< (/ (sum nlegal) (length nlegal)) .3) (newa (- A .3))))
) ;;end if

 ;; Calculate V's, PD
 (*when acte
   (*set cntx (!! 0))  (*set cnty (!! 0))  ;; To count interactions
   (*set PD (cond!!     ((and!! Vup Vdn) (!! (- A)))
                        ((xor!! Vup Vdn) (!! 0.0))
                        (t!! (!! A))))
    ) ;; end *when acte
;; this statement sets the INCRement and the initial value of EX 
(if alternate (progn (setq incr (- incr)) (setq ex (- ex incr)))
              (setq ex (- ex (* Ncl incr))))
;; Loop through edge classes
(dotimes (dmb Ncl)	;; dmb = 'dumb' variable is not used in loop
    (setq ex (+ ex incr))
    ;; Calculate Up/Down  interactions
    (*set Vud (!! 0))
    (*if Vdn (*set Vud (!! 1)))
    (*if Vup (*set Vud (logior!! Vud (!! 2))))
    (*with-vp-set vpinter
      (*when intrp
       (*set iactv nil!!)
       (*when (=!! clab (!! ex))
	 (*set iactv t!!) (*set isrc rpcd) (*set itgt rpab))
       (*when (=!! clcd (!! ex))
	 (*set iactv t!!) (*set isrc rpab) (*set itgt rpcd))
       (*when iactv 
         (*set Iud (pref!! Vud isrc)) 
           (*if (logbitp!! (!! 1) Iud)
		(CM:send-with-s-add-1L cntx-addr itgt-addr ONE-addr
				                 24 CM:*no-field*))
             ;;  ( (*pset :add (!! 1) cntx itgt :combine-with-dest t
	     ;;	                             :vp-set *default-vp-set*))
           (*if (logbitp!! (!! 0) Iud)
		(CM:send-with-s-add-1L cnty-addr itgt-addr ONE-addr
				                 24 CM:*no-field*))
             ;;  ( (*pset :add (!! 1) cnty itgt :combine-with-dest t
             ;;                                :vp-set *default-vp-set*))
)) ;; end *when iactv, end *when intrp
) ;; end *with-vp-set
 ;; Calculate new Down activations
  (*when acte (*when (=!! edgc (!! ex))
    (*set Du (-!! PD (*!! (!! B) cnty)))
    (*set Udn (+!! Udn (*!! Du (!! DT))))
 ;; Calculate new Up activations
    (*set Du (-!! PD (*!! (!! B) cntx)))
    (*set Uup (+!! Uup (*!! Du (!! DT))))
 ;; The following expression prevents introduction of up/down conflicts:
      (*when (and!! (plusp!! Uup) (plusp!! Udn))
        (*if (>=!! Uup Udn) (*set Udn (-!! Udn (!! A*DT)))
                            (*set Uup (-!! Uup (!! A*DT)))))
    (*set Vup (plusp!! Uup)) (*set Vdn (plusp!! Udn))
 )) ;; end *when acte, end *when (=!! edgc (!! ex))
) ;; end dotimes
 (*when acte  ;; To count non-trivial edges
    (*when (or!! Vup Vdn) (setq ntredges (*sum (!! 1))))
 ) ;; end when acte
 (incf cycles)
 (setq totedges (+ trvedges ntredges))   ;; Number of edges now
 (testplanar Vup Vdn)                    ;; Test for planarity, maximality
 (when (and legal (> totedges maxedges)) ;; Remember best legal configuration
   (setq maxedges totedges) (*set Vxp Vup) (*set Vxn Vdn))
 ;; graphics saver
(svpresnt "edges.list")
 ;; for the self-adjust mechanism
(setq nlegal (append nlegal (list (b01 legal))))
(if (< 7 (length nlegal)) (setq nlegal (cdr nlegal)))
(list 'Legal legal 'More more 'Cycles cycles 'Total-edges totedges 'A a)
)

;; Test for planarity (including up/down conflicts) and maximality
;; Call from default vp-set with all processors active.
;; Sets `legal' to (graph is planar and has no up/down conflicts)
;; Sets `more' to number of edges which are not present but could
;;      (individually) be incorporated without conflict, >= 0.

(*proclaim '(type (unsigned-pvar 2) Zud))
(*defvar Zud (!! 0))

(*defun testplanar (Vp Vn)
  (declare (type Boolean-pvar Vp Vn))
;; Calculate conflicts
 (*set Zud (!! 0))
 (*set Vud (!! 0))
 (*if Vn (*set Vud (!! 1)))
 (*if Vp (*set Vud (logior!! Vud (!! 2))))
 (*with-vp-set vpinter
    (*when intrp
      (*set Iud (pref!! Vud rpab))
      (*if (logbitp!! (!! 0) Iud) 
	 (*pset :logior (!! 1) Zud rpcd
		:vp-set *default-vp-set* :combine-with-dest t))
      (*if (logbitp!! (!! 1) Iud)
	   (*pset :logior (!! 2) Zud rpcd 
		:vp-set *default-vp-set* :combine-with-dest t))
      (*set Iud (pref!! Vud rpcd))
      (*if (logbitp!! (!! 0) Iud)
	   (*pset :logior (!! 1) Zud rpab
		:vp-set *default-vp-set* :combine-with-dest t))
      (*if (logbitp!! (!! 1) Iud)
	   (*pset :logior (!! 2) Zud rpab 
		:vp-set *default-vp-set* :combine-with-dest t))
 )) ; end *when intrp, end *with-vp-set
  (*when acte
    (setq legal
          (not (*or (or!! (plusp!! (logand!! Vud Zud)) (and!! Vp Vn)))))
    (*when (and!! (not!! Vp) (not!! Vn) (>!! (!! 3) Zud))
      (setq more (*sum (!! 1))))
  ) ; end *when acte
)


;; Install new value for A

(defun newa (av) (setq A av) (setq A*DT (* A DT)) (setq nlegal ())
                 (list A A*DT))

